import bpy
from bpy.types import (
    PropertyGroup,
    Object,
)
from bpy.props import (
    BoolProperty,
    IntProperty,
    FloatProperty,
    FloatVectorProperty,
    PointerProperty,
    StringProperty,
    EnumProperty,
    CollectionProperty,
)

import os
import json

from datetime import date

from . waves import (
    update_waves,
    M_PI,
)

from . ocean import (
    update_wakes,
    update_camera,
    update_ocean,
    update_shrinkwrap,
    update_world_volumetrics,
    update_world_underwater,
    update_submerged,
    update_intersection,
)

from . preferences import get_addon_version


# PRESET FUNCTIONS

def validate_preset_name(preset_name):
    # A list of symbols that cannot be accepted in a file name
    invalid_symbols = "\\/?:*|\"<>"
    for i in preset_name:
        if i in invalid_symbols:
            return ("Invalid symbols found from the following: %s" % invalid_symbols, invalid_symbols)
    # Check if name is empty or all spaces
    if len(preset_name.replace(" ", "")) == 0:
        return ("Preset name cannot be empty", invalid_symbols)
    return None


def is_serializable(x):
    """A function to check if a param is JSON seriazible"""
    try:
        json.dumps(x)
        return True
    except:
        return False


def create_preset(context, preset_path):
    """Function that writes a preset based on current POW ocean to a .json fila given a path"""

    world = context.scene.world

    preset_data = {"ocean": {}, "version": get_addon_version()} # TODO Find a way to automatically get current version

    for prop_name, prop in world.pow_settings.ocean.pow_ocean.items():
        # Check if prop is serializable to avoid writing object pointers in presets
        # Color and Vector types don't return as serializable, but still work
        if is_serializable(prop):
            preset_data["ocean"][prop_name] = getattr(world.pow_settings.ocean.pow_ocean, prop_name)
        else:
            # Assume prop is iterable
            # FIXME if in future more non-serializable types appear than just vectors/arrays
            write_prop = [] 
            for element in prop:
                write_prop.append(element)
            preset_data["ocean"][prop_name] = write_prop

    with open(preset_path, 'w') as json_file:
        json.dump(preset_data, json_file, indent=4)


def backup_preset(preset_path, preset_data, version):
    """Function that backups a preset given it's path and version data"""
    dir_path = os.path.split(preset_path)[0] # Get the preset's directory
    backup_path = os.path.join(dir_path, "backup_%s_%s" % (str(version), date.today()))
    os.makedirs(backup_path, exist_ok=True)
    backup_preset_path = os.path.join(backup_path, os.path.split(preset_path)[1])

    with open(backup_preset_path, 'w') as json_file:
        json.dump(preset_data, json_file, indent=4)
    

def update_preset_versions(context):
    pow_presets_dir = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")

    for file in os.listdir(pow_presets_dir):
        file_name = os.fsdecode(file)
        if file_name.endswith(".json"):
            preset_path = os.path.join(pow_presets_dir, file_name)
            with open(preset_path) as json_data:
                data = json.load(json_data)
                ocean = data.get("ocean")

                if ocean.get("version") == None:
                    # First version of presets with no metadata, versions 1.X
                    backup_preset(preset_path, data, (1, 0, 0))
                    if ocean.get("water_color"):
                        # Automatically calculate logical extenction coefficient
                        ocean["extinction_coefficient"] = (1.0 - ocean["water_color"][0], 1.0 - ocean["water_color"][1], 1.0 - ocean["water_color"][2])
                    if ocean.get("particles"):
                        ocean["particles"] = 'enabled' if ocean["particles"] else 'disabled'
                    if ocean.get("water_edge"):
                        ocean["water_edge"] = 'enabled' if ocean["water_edge"] else 'disabled'
                    data["version"] = get_addon_version() # Update to current version, TODO auto obtain this version number somehow
                    
                    with open(preset_path, 'w') as json_file:
                        json.dump(data, json_file, indent=4)


def apply_preset(context, preset_path):
    """Function that loads a preset .json from given path and applies to current POW ocean"""

    world = context.scene.world

    with open(preset_path) as json_data:
        data = json.load(json_data)
    
    ocean = data.get("ocean")

    # Register the class to set defaults to the preset
    new_ocean = register_pow_property_groups(context, ocean)

    if new_ocean:
        for prop_name in new_ocean.keys():
            setattr(world.pow_settings.ocean.pow_ocean, prop_name, new_ocean.get(prop_name))


def select_preset(self, context):
    world = context.scene.world

    # Check to ensure the preset index is valid
    if len(world.pow_settings.presets) > world.pow_settings.preset_index and world.pow_settings.preset_index >= 0:
        preset_data = world.pow_settings.presets[world.pow_settings.preset_index]
        apply_preset(context, preset_data.path) # Apply preset
        world.pow_settings.preset_name = preset_data.name


def update_preset_list(self, context):
    world = context.scene.world

    # Update presets list
    world.pow_settings.presets.clear()

    # Find the directory of our presets and parse JSONs from there
    preset_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
    os.makedirs(preset_path, exist_ok=True)

    for file in os.listdir(preset_path):
        file_name = os.fsdecode(file)
        if file_name.endswith(".json"):
            preset_data = world.pow_settings.presets.add()
            preset_data["name"] = file_name.removesuffix(".json")
            preset_data.path = os.path.join(preset_path, file_name)


def rename_preset(self, context):
    world = context.scene.world

    # We need the previous name to find the file and rename it
    old_name = world.pow_settings.preset_name
    new_name = world.pow_settings.presets[world.pow_settings.preset_index].name

    # Generate the new file path for renaming
    base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")

    old_path = os.path.join(base_path, old_name + ".json")
    new_path = os.path.join(base_path, new_name + ".json")

    # Check if the name contains invalid symbols
    invalid_symbols = validate_preset_name(new_name)
    if invalid_symbols:
        world.pow_settings.presets[world.pow_settings.preset_index].name = old_name
        # TODO See if user can be notified of this operation's failure somehow
    # Check if we're not clashing with existing preset name
    elif os.path.isfile(new_path):
        world.pow_settings.presets[world.pow_settings.preset_index].name = old_name
        # TODO See if user can be notified of this operation's failure somehow
    else:
        # Rename the file
        os.rename(old_path, new_path)

        # Update preset list to ensure the new name is shown up in UI
        update_preset_list(self, context)

        # Make sure the renamed preset is still selected
        for i in range(len(world.pow_settings.presets)):
            if world.pow_settings.presets[i].name == new_name:
                world.pow_settings.preset_index = i
                break


class POW_Preset(PropertyGroup):
    """POW Preset Object"""

    name: StringProperty(
        name = "Name",
        description = "Name of preset",
        update = rename_preset,
    )

    path: StringProperty(
        subtype = "FILE_PATH",
        name = "File Path",
        description = "Path of the JSON file that contains preset data",
    )


# POW PROPERTY GROUPS

class POW_Settings(PropertyGroup):
    """POW Settings"""
    
    ocean: PointerProperty(
        name = "Ocean Object",
        description = "Ocean object",
        type = Object,
    )

    camera: PointerProperty(
        name = "Camera",
        description = "Current camera object used by addon's LODs",
        type = Object,
        update = update_camera,
    )

    use_active_camera: BoolProperty(
        name = "Use Active Camera",
        description = "Always use current active camera object for addon's LOD calculation (currently unimplemented)",
        default = True,
        # TODO Implement this.
    )

    sun: PointerProperty(
        name = "Sun",
        description = "Sun object used for scattering calculations",
        type = Object,
        update = update_ocean,
    )

    wake_object: PointerProperty(
        name = "Wake Object",
        description = "Object that describes the position and shape of wakes",
        type = Object,
        update = update_wakes,
    )

    shrinkwrap_object: PointerProperty(
        name = "Shrinkwrap Object",
        description = "Object that ocean surface is wrapped to. Use for custom wave shapes in an area.",
        type = Object,
        update = update_shrinkwrap,
    )

    world_underwater: EnumProperty(
        name = "World Underwater",
        description = "Used to produce correct reflection/refraction of water in Eevee and to reduce potential artifacts.",
        items=[(
                'enabled', 
                'Enabled', 
                'Enable world underwater volumetrics'
            ),
            (
                'disabled', 
                'Disabled', 
                'Disable world underwater volumetrics'
            )],
        default = 'disabled',
        update=update_world_underwater,
    )

    world_volumetrics: EnumProperty(
        name = "World Volumetrics",
        description = "Advanced Blender underwater volumetrics that comes with shadows and interaction with lights, but scattering algorithm is not as advanced for water. These are applied on top of existing volumetrics and can have a performance impact.",
        items=[(
                'enabled', 
                'Enabled',
                'Enable world underwater volumetrics'
            ),
            (
                'disabled', 
                'Disabled', 
                'Disable world underwater volumetrics'
            )],
        default = 'disabled',
        update=update_world_volumetrics,
    )

    submerged_shader: BoolProperty(
        name = "Submerged Shader",
        description = "Submerged object underwater shader",
        default = False,
        update = update_submerged,
    )

    submerged_advanced_intersection: BoolProperty(
        name = "Advanced Intersection",
        description = "Advanced intersection for submerged shader",
        default = False,
        update = update_intersection,
    )

    presets: CollectionProperty(
        type = POW_Preset,
        name = "Presets",
        description = "A list of all available presets for POW",
    )

    preset_name: StringProperty(
        name = "Preset Name",
        description = "Name of current POW preset",
        default = '',
    )

    preset_index: IntProperty(
        name = "Preset Index",
        description = "Index of preset selected to apply",
        default = -1,
        update = select_preset,
    )

    show_presets: BoolProperty(
        name = "Show Presets",
        description = "Show Presets in UI",
        default = False,
        update = update_preset_list,
    )


def register_pow_property_groups(context, ocean, first_time=False):
    """Function that sets default values for all props and re-registers them."""

    ocean_defaults = {
        "material_type": 'refractive',
        "water_color": (0.02, 0.078, 0.124),
        "extinction_coefficient": (0.98, 0.922, 0.876),
        "scattering_distance": 0.2,
        "vertical_scattering": 12.0,
        "vertical_scattering_exponent": 6.0,
        "snell_scattering": 12.0,
        "forward_scattering": 2.0,
        "back_scattering": 0.0,
        "camera_near_offset": 0.02,
        "volume_density": 0.1,
        "volume_distance": 100.0,
        "water_subsurface_scale": 32.0,
        "foam_amount": 0.12,
        "lambda_min": 0.2,
        "lambda_max": 50.0,
        "height_max": 1.0,
        "wave_distribution_log": 2.0,
        "wave_dispersion": M_PI,
        "wave_angle": M_PI / 2.0,
        "wind_speed": 10.0,
        "time_scale": 1.0,
        "wave_choppiness": 0.15,
        "wave_skew": -0.2,
        "base_size": 1000,
        "base_resolution": 2,
        "detail_resolution": 3,
        "view_frustum_boost": 2.0,
        "far_lod_offset": 0.85,
        "near_lod_offset": 0.2,
        "lod_exponent": 3.0,
        "wake_height": 0.0,
        "capillary_strength": 0.02,
        "capillary_scale": 3.0,
        "capillary_stretch": 3.0,
        "capillary_direction": M_PI,
        "capillary_speed": 1.0,
        "capillary_variance": 1.0,
        "capillary_drag": 0.7,
        "capillary_mask_scale": 0.001,
        "capillary_mask_minimum": 0.4,
        "capillary_mask_maximum": 0.6,
        "capillary_roughness_minimum": 0.01,
        "capillary_roughness_maximum": 0.1,
        "particles": 'disabled',
        "particle_domain_size": 10.0,
        "particle_domain_z_offset": 0.05,
        "water_edge": 'disabled',
        "water_edge_z_offset": 0.001,
        "water_edge_width": 1.0,
        "water_edge_height": 0.01,
        "water_edge_resolution": 8,
    }

    # Replace all the default preset values with custom ones
    new_ocean_defaults = ocean_defaults.copy()
    for key, value in ocean.items():
        new_ocean_defaults[key] = value

    class POW_Ocean(PropertyGroup):
        """POW Ocean Instance"""

        # MATERIAL PROPERTIES

        material_type: EnumProperty(
            name = "Material Type",
            description = "Which shader model material uses",
            items=[(
                'refractive', 
                'Refractive', 
                'Transparent refractive ocean material'
            ),
            (
                'subsurface', 
                'Subsurface Scatter', 
                'Subsurface scattering approximation of ocean material'
            )
            ],
            default = new_ocean_defaults["material_type"],
            update = update_ocean,
        )

        water_color: FloatVectorProperty(
            name = "Color",
            description = "Water subsurface color",
            subtype = 'COLOR',
            size = 3,
            min = 0.0,
            max = 1.0,
            default = new_ocean_defaults["water_color"],
        )

        extinction_coefficient: FloatVectorProperty(
            name = "Extinction",
            description = "Water extinction",
            subtype = 'COLOR',
            size = 3,
            min = 0.0,
            max = 1.0,
            default = new_ocean_defaults["extinction_coefficient"],
        )

        scattering_distance: FloatProperty(
            name = "Scattering Distance",
            description = "Scattering distance multiplier for extinction coefficient",
            min = 0.0,
            soft_max = 10.0,
            default = new_ocean_defaults["scattering_distance"],
        )

        vertical_scattering: FloatProperty(
            name = "Vertical Scattering",
            description = "Vertical scattering amount",
            min = 0.0,
            soft_max = 20.0,
            default = new_ocean_defaults["vertical_scattering"],
        )

        vertical_scattering_exponent: FloatProperty(
            name = "Vertical Scattering Exponent",
            description = "Vertical scattering exponent",
            min = 0.0,
            soft_min = 1.0,
            soft_max = 10.0,
            default = new_ocean_defaults["vertical_scattering_exponent"],
        )

        snell_scattering: FloatProperty(
            name = "Snell's Scattering",
            description = "Snell's scattering amount",
            min = 0.0,
            soft_max = 20.0,
            default = new_ocean_defaults["snell_scattering"],
        )

        forward_scattering: FloatProperty(
            name = "Forward Scattering",
            description = "Forward scattering amount",
            min = 0.0,
            soft_max = 5.0,
            default = new_ocean_defaults["forward_scattering"],
        )

        back_scattering: FloatProperty(
            name = "Back Scattering",
            description = "Back scattering amount",
            min = 0.0,
            soft_max = 5.0,
            default = new_ocean_defaults["back_scattering"],
        )

        camera_near_offset: FloatProperty(
            name = "Camera Near Offset",
            description = "Offset value for camera near plane (..)",
            subtype = 'FACTOR',
            min = 0.0,
            soft_max = 0.05,
            default = new_ocean_defaults["camera_near_offset"],
        )

        volume_density: FloatProperty(
            name = "Volume Density",
            description = "Density multiplier of world volumetric shader",
            min = 0.0,
            soft_max = 2.0,
            default = new_ocean_defaults["volume_density"],
        )

        volume_distance: FloatProperty(
            name = "Volume Distance",
            description = "Distance range of world volumetric shader",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_max = 100.0,
            default = new_ocean_defaults["volume_distance"],
        )

        water_subsurface_scale: FloatProperty(
            name = "Subsurface Scale",
            description = "Water subsurface scale",
            min = 0.0,
            soft_max = 100.0,
            default = new_ocean_defaults["water_subsurface_scale"],
        )

        foam_amount: FloatProperty(
            name = "Foam Amount",
            description = "Foam amount",
            subtype = 'FACTOR',
            min = 0.0,
            soft_max = 0.5,
            default = new_ocean_defaults["foam_amount"],
        )
        
        # PYTHON-SIDE WAVE CALCULATION PARAMS

        lambda_min: FloatProperty(
            name = "Smallest Wave",
            description = "Minimum wave frequency",
            subtype = 'DISTANCE',
            min = 0.000001,
            soft_max = 10.0,
            default = new_ocean_defaults["lambda_min"],
            update = update_waves,
        )

        lambda_max: FloatProperty(
            name = "Largest Wave",
            description = "Maximum wave frequency",
            subtype = 'DISTANCE',
            min = 0.000001,
            soft_max = 500.0,
            default = new_ocean_defaults["lambda_max"],
            update = update_waves,
        )

        height_max: FloatProperty(
            name = "Wave Height",
            description = "Maximum wave amplitude",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_max = 10.0,
            default = new_ocean_defaults["height_max"],
            update = update_waves,
        )

        wave_distribution_log: FloatProperty(
            name = "Distribution Log",
            description = "Wave distribution log",
            subtype = 'FACTOR',
            min = 0.0000001,
            soft_min = 1.05,
            soft_max = 5.0,
            default = new_ocean_defaults["wave_distribution_log"],
            update = update_waves,
        )

        wave_dispersion: FloatProperty(
            name = "Dispersion",
            description = "Wave random rotation amount",
            subtype = 'ANGLE',
            default = new_ocean_defaults["wave_dispersion"],
            update = update_waves,
        )

        wave_angle: FloatProperty(
            name = "Angle",
            description = "Angle of wave direction",
            subtype = 'ANGLE',
            default = new_ocean_defaults["wave_angle"],
            update = update_waves,
        )
        
        wind_speed: FloatProperty(
            name = "Wind Speed",
            description = "Wind speed at ground level",
            unit = 'VELOCITY',
            min = 0.0,
            soft_max = 30.0,
            default = new_ocean_defaults["wind_speed"],
            update = update_waves,
        )

        # GN SIDE WAVE PARAMS

        time_scale: FloatProperty(
            name = "Time Multiplier",
            description = "Use to control how fast waves move in simulation",
            subtype = 'FACTOR',
            min = 0,
            soft_min = 0.2,
            soft_max = 5.0,
            default = new_ocean_defaults["time_scale"],
            #update = update_ocean,
        )

        wave_choppiness: FloatProperty(
            name = "Choppiness",
            description = "Control how choppy the waves appear",
            subtype = 'FACTOR',
            min = 0,
            soft_min = 0.0,
            soft_max = 0.3,
            default = new_ocean_defaults["wave_choppiness"],
            #update = update_ocean,
        )

        wave_skew: FloatProperty(
            name = "Skew",
            description = "Control wave skew towards concentric formations",
            subtype = 'FACTOR',
            soft_min = -2.0,
            soft_max = 2.0,
            default = new_ocean_defaults["wave_skew"],
            #update = update_ocean,
        )

        # MESH SETTINGS

        base_size: FloatProperty(
            name = "Size",
            description = "Ocean mesh size",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_min = 1.0,
            soft_max = 5000.0,
            default = new_ocean_defaults["base_size"],
            #update = update_ocean,
        )

        base_resolution: IntProperty(
            name = "Resolution",
            description = "Ocean mesh resolution",
            min = 0,
            soft_min = 1,
            soft_max = 5,
            default = new_ocean_defaults["base_resolution"],
            #update = update_ocean,
        )

        detail_resolution: IntProperty(
            name = "Detail Resolution",
            description = "Ocean mesh detail resolution",
            min = 0,
            soft_min = 1,
            soft_max = 10,
            default = new_ocean_defaults["detail_resolution"],
            #update = update_ocean,
        )

        # CAMERA SETTINGS

        view_frustum_boost: FloatProperty(
            name = "View Frustum Boost",
            description = "Boost view frustum radius to reduce potential gaps on visible ocean",
            subtype = 'FACTOR',
            min = 0,
            soft_min = 1.0,
            soft_max = 3.0,
            default = new_ocean_defaults["view_frustum_boost"],
            #update = update_ocean,
        )

        far_lod_offset: FloatProperty(
            name = "Far LOD Offset",
            description = "Far LOD offset",
            subtype = 'FACTOR',
            soft_min = 0.3,
            soft_max = 1.3,
            default = new_ocean_defaults["far_lod_offset"],
            #update = update_ocean,
        )

        near_lod_offset: FloatProperty(
            name = "Near LOD Offset",
            description = "Near LOD offset",
            subtype = 'FACTOR',
            soft_min = -0.3,
            soft_max = 0.7,
            default = new_ocean_defaults["near_lod_offset"],
            #update = update_ocean,
        )

        lod_exponent: FloatProperty(
            name = "LOD Exponent",
            description = "Exponent for LOD size based on distance",
            subtype = 'FACTOR',
            min = 0,
            soft_min = 1.0,
            soft_max = 5.0,
            default = new_ocean_defaults["lod_exponent"],
            #update = update_ocean,
        )

        # WAKE SETTINGS

        wake_height: FloatProperty(
            name = "Wake Height",
            description = "Height of the wakes in ocean",
            subtype = 'DISTANCE',
            min = 0,
            soft_max = 5.0,
            default = new_ocean_defaults["wake_height"],
            #update = update_ocean,
        )

        # CAPILLARY WAVE SETTINGS

        capillary_strength: FloatProperty(
            name = "Strength",
            description = "Strength of the capillary waves",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 0.2,
            default = new_ocean_defaults["capillary_strength"],
            #update = update_ocean,
        )

        capillary_scale: FloatProperty(
            name = "Scale",
            description = "How many capillary waves are present per area",
            min = 0,
            soft_max = 10.0,
            default = new_ocean_defaults["capillary_scale"],
            #update = update_ocean,
        )

        capillary_stretch: FloatProperty(
            name = "Stretch",
            description = "Stretching of the capillary waves by the wind",
            min = 0,
            soft_max = 10.0,
            default = new_ocean_defaults["capillary_stretch"],
            #update = update_ocean,
        )

        capillary_direction: FloatProperty(
            name = "Direction",
            description = "Direction of the capillary waves",
            subtype = 'ANGLE',
            soft_min = 0,
            soft_max = 2.0 * M_PI,
            default = new_ocean_defaults["capillary_direction"],
            #update = update_ocean,
        )

        capillary_speed: FloatProperty(
            name = "Speed",
            description = "Speed of the capillary waves",
            #unit = 'VELOCITY', # TODO verify if this is a m/s value
            soft_min = 0,
            soft_max = 10.0,
            default = new_ocean_defaults["capillary_speed"],
            #update = update_ocean,
        )

        capillary_variance: FloatProperty(
            name = "Variance",
            description = "Variance of the capillary waves",
            subtype = 'FACTOR',
            soft_min = 0,
            soft_max = 2.0,
            default = new_ocean_defaults["capillary_variance"],
            #update = update_ocean,
        )

        capillary_drag: FloatProperty(
            name = "Drag",
            description = "Drag of the capillary waves, affects their apparent viscosity",
            soft_min = 0,
            soft_max = 2.0,
            default = new_ocean_defaults["capillary_drag"],
            #update = update_ocean,
        )

        capillary_mask_scale: FloatProperty(
            name = "Mask Scale",
            description = "Scale of the detail mask of the capillary waves",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 0.02,
            default = new_ocean_defaults["capillary_mask_scale"],
            #update = update_ocean,
        )

        capillary_mask_minimum: FloatProperty(
            name = "Mask Minimum",
            description = "Minimum factor value of the capillary wave mask",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 1.0,
            default = new_ocean_defaults["capillary_mask_minimum"],
            #update = update_ocean,
        )

        capillary_mask_maximum: FloatProperty(
            name = "Mask Maximum",
            description = "Maximum factor value of the capillary wave mask",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 1.0,
            default = new_ocean_defaults["capillary_mask_maximum"],
            #update = update_ocean,
        )

        capillary_roughness_minimum: FloatProperty(
            name = "Roughness Minimum",
            description = "Minimum roughness of the capillary wave",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 1.0,
            default = new_ocean_defaults["capillary_roughness_minimum"],
            #update = update_ocean,
        )

        capillary_roughness_maximum: FloatProperty(
            name = "Roughness Maximum",
            description = "Maximum roughness of the capillary wave",
            subtype = 'FACTOR',
            min = 0,
            soft_max = 1.0,
            default = new_ocean_defaults["capillary_roughness_maximum"],
            #update = update_ocean,
        )

        particles: EnumProperty(
            name = "Particles",
            description = "Underwater particle effect",
            items=[(
                'enabled', 
                'Enabled', 
                'Enable underwater particles'
            ),
            (
                'disabled', 
                'Disabled', 
                'Disable underwater particles'
            )
            ],
            default = new_ocean_defaults["particles"],
            update = update_ocean,
        )

        particle_domain_size: FloatProperty(
            name = "Domain Size",
            description = "Particle domain size",
            subtype = 'DISTANCE',
            min = 0.1,
            soft_max = 20.0,
            default = new_ocean_defaults["particle_domain_size"],
            #update = update_ocean,
        )
        particle_domain_z_offset: FloatProperty(
            name = "Domain Z Offset",
            description = "Particle domain Z offset",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_max = 0.5,
            default = new_ocean_defaults["particle_domain_z_offset"],
            #update = update_ocean,
        )

        water_edge: EnumProperty(
            name = "Water Edge",
            description = "Water contact edge with camera",
            items=[(
                'enabled', 
                'Enabled', 
                'Enable water/camera intersection shader'
            ),
            (
                'disabled', 
                'Disabled', 
                'Disable water/camera intersection shader'
            )
            ],
            default = new_ocean_defaults["water_edge"],
            update = update_ocean,
        )

        water_edge_z_offset: FloatProperty(
            name = "Z Offset",
            description = "Water edge Z offset",
            subtype = 'DISTANCE',
            soft_min = -0.1,
            soft_max = 0.1,
            default = new_ocean_defaults["water_edge_z_offset"],
            #update = update_ocean,
        )
        
        water_edge_width: FloatProperty(
            name = "Width",
            description = "Water edge width",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_max = 3.0,
            default = new_ocean_defaults["water_edge_width"],
            #update = update_ocean,
        )
        
        water_edge_height: FloatProperty(
            name = "Height",
            description = "Water edge height",
            subtype = 'DISTANCE',
            min = 0.0,
            soft_max = 0.2,
            default = new_ocean_defaults["water_edge_height"],
            #update = update_ocean,
        )

        water_edge_resolution: IntProperty(
            name = "Resolution",
            description = "Water edge resolution",
            min = 2,
            soft_max = 32,
            default = new_ocean_defaults["water_edge_resolution"],
            #update = update_ocean,
        )
    
    bpy.utils.register_class(POW_Ocean)

    bpy.types.Object.pow_ocean = PointerProperty(type = POW_Ocean)

    return (new_ocean_defaults)
