import bpy
from bpy.types import (
    Panel,
    UIList,
    Operator,
)

import os, sys, subprocess, platform

from . properties import (
    create_preset,
    update_preset_list,
    validate_preset_name,
)

from . ocean import (
    update_underwater,
)


class DrawablePanel:
    """可拉伸POW面板的通用类别"""

    @classmethod
    def poll(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean

        if ocean:
            return True
        return False

    def draw_header(self, context):
        self.layout.label(text='', icon=self.icon_name)

    # Indented column with a padding from the left.
    def indented_column(self, squished=True):
        row = self.layout.row()
        row.separator()
        return row.column(align=squished)
    

class POW_TB:
    """工具栏元素的基类"""
    bl_parent_id = "POW_PT_OceanSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"


class OceanSettings(DrawablePanel):
    bl_label = "POW Ocean"

    @classmethod
    def poll(self, context):
        # Override poll method since this header should always be present.
        return True

    def draw_header(self, context):
        pass
    
    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.layout

        # MacOS warning and easily accessible graphics API switch
        if platform.system() == 'Darwin':
            if bpy.context.preferences.system.gpu_backend == 'METAL':
                box = layout.box()
                box.label(
                    text="警告Metal API具有节点限制，可能会导致POW无法在MacOS上工作。尝试切换到OpenGL（如果可用）：",
                    icon='ERROR')
                box.prop(bpy.context.preferences.system, "gpu_backend")

        if not ocean:
            layout.operator("pow.add_ocean", icon = "ADD")
            #layout.prop(world.pow_settings, "ocean")
        if ocean:
            # Presets
            # Preset selection UI
            col = layout.column()
            col.label(text="Presets:")
            if world.pow_settings.show_presets:
                preset_text = "Select Preset"
                if world.pow_settings.preset_index >= 0:
                    preset_text = world.pow_settings.presets[world.pow_settings.preset_index].name
                col.prop(world.pow_settings, "show_presets", icon="DOWNARROW_HLT", text=preset_text)
                box = col.box()
                box.template_list("POW_UL_PresetList", "", context.scene.world.pow_settings, "presets", context.scene.world.pow_settings, "preset_index", rows=6)
            else:
                preset_text = "Select Preset"
                if world.pow_settings.preset_index >= 0:
                    preset_text = world.pow_settings.presets[world.pow_settings.preset_index].name
                col.prop(world.pow_settings, "show_presets", icon="RIGHTARROW_THIN", text=preset_text)

            box = col.box()
            row = box.row()
            row.enabled = False
            if world.pow_settings.preset_index >= 0:
                row.enabled = True # Only enable this row of controls when a non-default preset is selected
            row.operator("pow.save_current_preset", icon="FILE_TICK", text="Save")
            row.operator("pow.delete_current_preset", icon="TRASH", text="")
            row = box.row()
            row.operator("pow.create_new_preset", icon="ADD")
            row.operator("pow.open_presets_folder", icon="FILE_FOLDER", text="")

            # Sun
            col.prop(world.pow_settings, "sun")


class POW_PT_OceanSettingsTB(Panel, OceanSettings):
    """Base class for placing all the UI elements in world tab"""
    bl_idname = "POW_PT_OceanSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"


class MeshSettings(DrawablePanel):
    bl_label = "Mesh"
    icon_name = "MESH_GRID"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.prop(ocean, "location", index=2, text="Height")
        layout.separator()

        layout.prop(ocean.pow_ocean, "base_size")
        layout.prop(ocean.pow_ocean, "base_resolution")
        layout.prop(ocean.pow_ocean, "detail_resolution")

        layout.label(text="Shrinkwrap:")
        layout.prop(world.pow_settings, "shrinkwrap_object", text="")


class POW_PT_MeshSettingsSettingsTB(POW_TB, Panel, MeshSettings):
    """Mesh Settings UI Class"""
    bl_idname = "POW_PT_MeshSettingsSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class MaterialSettings(DrawablePanel):
    bl_label = "Material"
    icon_name = "MATERIAL"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        row = layout.row()
        row.prop(ocean.pow_ocean, "material_type", expand=True)

        layout.prop(ocean.pow_ocean, "water_color")
        if ocean.pow_ocean.material_type == 'subsurface':
            layout.prop(ocean.pow_ocean, "water_subsurface_scale")
        else:
            layout.prop(ocean.pow_ocean, "extinction_coefficient")
            layout.prop(ocean.pow_ocean, "scattering_distance")
            layout.prop(ocean.pow_ocean, "vertical_scattering")
            layout.prop(ocean.pow_ocean, "vertical_scattering_exponent")
            layout.prop(ocean.pow_ocean, "snell_scattering")
            layout.prop(ocean.pow_ocean, "forward_scattering")
            layout.prop(ocean.pow_ocean, "back_scattering")
            layout.prop(ocean.pow_ocean, "camera_near_offset")


class POW_PT_MaterialSettingsTB(POW_TB, Panel, MaterialSettings):
    """Material Settings UI Class"""
    bl_idname = "POW_PT_MaterialSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class CameraSettings(DrawablePanel):
    bl_label = "Camera & LOD"
    icon_name = "VIEW_CAMERA"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.prop(world.pow_settings, "camera")
        if world.pow_settings.camera:
            layout.prop(world.pow_settings.camera.data, "clip_end", text="End")
        layout.separator()
        layout.prop(ocean.pow_ocean, "view_frustum_boost")
        layout.prop(ocean.pow_ocean, "far_lod_offset")
        layout.prop(ocean.pow_ocean, "near_lod_offset")
        layout.prop(ocean.pow_ocean, "lod_exponent")
        layout.separator()
        layout.prop(ocean.pow_ocean, "time_scale")

        layout.separator()
        layout.label(text="Water Edge:")
        row = layout.row()
        row.prop(ocean.pow_ocean, "water_edge", expand=True)
        if ocean.pow_ocean.water_edge in {'enabled'}:
            layout.prop(ocean.pow_ocean, "water_edge_z_offset")
            layout.prop(ocean.pow_ocean, "water_edge_width")
            layout.prop(ocean.pow_ocean, "water_edge_height")
            layout.prop(ocean.pow_ocean, "water_edge_resolution")


class POW_PT_CameraSettingsTB(POW_TB, Panel, CameraSettings):
    """Camera Settings UI Class"""
    bl_idname = "POW_PT_CameraSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class WavesSettings(DrawablePanel):
    bl_label = "Waves"
    icon_name = "MOD_OCEAN"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.prop(ocean.pow_ocean, "lambda_min")
        layout.prop(ocean.pow_ocean, "lambda_max")
        layout.prop(ocean.pow_ocean, "height_max")
        layout.prop(ocean.pow_ocean, "wind_speed")
        layout.separator()
        layout.prop(ocean.pow_ocean, "wave_distribution_log")
        layout.prop(ocean.pow_ocean, "wave_dispersion")
        layout.prop(ocean.pow_ocean, "wave_angle")
        layout.prop(ocean.pow_ocean, "wave_skew")
        layout.prop(ocean.pow_ocean, "wave_choppiness")
        layout.prop(ocean.pow_ocean, "foam_amount")


class POW_PT_WavesSettingsTB(POW_TB, Panel, WavesSettings):
    """Wave Settings UI Class"""
    bl_idname = "POW_PT_WavesSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class CapillarySettings(DrawablePanel):
    bl_label = "Capillary Waves"
    icon_name = "FORCE_FORCE"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.prop(ocean.pow_ocean, "capillary_strength")
        layout.prop(ocean.pow_ocean, "capillary_scale")
        layout.prop(ocean.pow_ocean, "capillary_stretch")
        layout.prop(ocean.pow_ocean, "capillary_direction")
        layout.prop(ocean.pow_ocean, "capillary_speed")
        layout.prop(ocean.pow_ocean, "capillary_variance")
        layout.prop(ocean.pow_ocean, "capillary_drag")
        layout.separator()
        layout.prop(ocean.pow_ocean, "capillary_mask_scale")
        layout.prop(ocean.pow_ocean, "capillary_mask_minimum")
        layout.prop(ocean.pow_ocean, "capillary_mask_maximum")
        layout.prop(ocean.pow_ocean, "capillary_roughness_minimum")
        layout.prop(ocean.pow_ocean, "capillary_roughness_maximum")


class POW_PT_CapillarySettingsTB(POW_TB, Panel, CapillarySettings):
    """Capillary Wave Settings UI Class"""
    bl_idname = "POW_PT_CapillarySettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class UnderwaterSettings(DrawablePanel):
    bl_label = "Underwater"
    icon_name = "POINTCLOUD_DATA"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.label(text="Objects:")
        col = layout.column()

        selected_materiables = 0
        if len(context.selected_objects) > 0:
            for object in context.selected_objects:
                if hasattr(object.data, "materials") and len(object.data.materials) > 0:
                    selected_materiables += 1
        if selected_materiables > 0:
            o_text = "objects"
            if selected_materiables == 1:
                o_text = "object"
            label_text = "%s " % selected_materiables + o_text + " selected"
            col.label(text=label_text)
            col = layout.column()
            col.enabled = True
        else:
            col.label(text="Select objects to apply")
            col = layout.column()
            col.enabled = False
        col.prop(world.pow_settings, "submerged_shader", text="Submerged Underwater")

        col = layout.column()
        col.enabled = False
        if world.pow_settings.submerged_shader and len(context.selected_objects) > 0:
            col.enabled = True
        col.prop(world.pow_settings, "submerged_advanced_intersection")
        
        col.separator()
        row = col.row()
        #row.operator("pow.apply_underwater_all", text="Add to All", icon="ADD")
        row.operator("pow.remove_underwater_all", text="Remove from All", icon="REMOVE")
        layout.separator()

        layout.label(text="Particles:")
        row = layout.row()
        row.prop(ocean.pow_ocean, "particles", expand=True)
        if ocean.pow_ocean.particles == 'enabled':
            layout.prop(ocean.pow_ocean, "particle_domain_size")
            layout.prop(ocean.pow_ocean, "particle_domain_z_offset")
        
        layout.separator()
        layout.label(text="World Underwater:")
        row = layout.row()
        row.prop(world.pow_settings, "world_underwater", expand=True)
        
        # Only expose world underwater settings if world has underwater nodes
        if world.pow_settings.world_underwater == 'enabled':
            layout.label(text="Advanced Volumetrics:")
            row = layout.row()
            row.prop(world.pow_settings, "world_volumetrics", expand=True)
            if world.pow_settings.world_volumetrics == "enabled":
                layout.prop(ocean.pow_ocean, "volume_density")
                layout.prop(ocean.pow_ocean, "volume_distance")


class POW_PT_UnderwaterSettingsTB(POW_TB, Panel, UnderwaterSettings):
    """Underwater Settings UI Class"""
    bl_idname = "POW_PT_UnderwaterSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class WakeSettings(DrawablePanel):
    bl_label = "Wakes & Ripples"
    icon_name = "SHARPCURVE"

    def draw(self, context):
        scene = context.scene
        world = scene.world
        ocean = world.pow_settings.ocean
        
        layout = self.indented_column()

        layout.label(text="Wakes:")
        layout.prop(world.pow_settings, "wake_object", text="")
        if world.pow_settings.wake_object != None:
            layout.prop(ocean.pow_ocean, "wake_height")

        layout.separator()
        layout.label(text="Ripples/Booleans:")
        box = layout.box()
        box.prop(ocean.modifiers["POW_Get_Intersection"], '["Socket_2"]', text="")
        if ocean.modifiers["POW_Get_Intersection"]["Socket_2"] != None:
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_11"]', text="Use Ripples")
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_4"]', text="Use as Boolean")
        box = layout.box()
        box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_2"]', text="")
        if ocean.modifiers["POW_Get_Intersection"]["Input_2"] != None:
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_12"]', text="Use Ripples")
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_5"]', text="Use as Boolean")
        box = layout.box()
        box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_3"]', text="")
        if ocean.modifiers["POW_Get_Intersection"]["Input_3"] != None:
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_13"]', text="Use Ripples")
            box.prop(ocean.modifiers["POW_Get_Intersection"], '["Input_6"]', text="Use as Boolean")


class POW_PT_WakeSettingsTB(POW_TB, Panel, WakeSettings):
    """Wake Settings UI Class"""
    bl_idname = "POW_PT_WakeSettingsTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'DEFAULT_CLOSED'}


class Footer(DrawablePanel):
    bl_label = "Extra Settings"
    icon_name = "SETTINGS"

    def draw(self, context):
        world = context.scene.world
        ocean = world.pow_settings.ocean

        layout = self.layout
        if ocean:
            layout.operator("pow.remove_ocean", icon = "X")


class POW_PT_FooterTB(POW_TB, Panel, Footer):
    """Footer UI Class"""
    bl_idname = "POW_PT_FooterTB"
    bl_region_type = "UI"
    bl_space_type = "VIEW_3D"
    bl_context = "objectmode"
    bl_category = "Ocean"
    bl_options = {'HIDE_HEADER'}

# PRESET CLASSES

class POW_UL_PresetList(UIList):
    """A list class for displaying presets"""

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        custom_icon = 'NONE' #'WORLD_DATA'

        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            layout.prop(item, "name", text="", expand=False, emboss=False, icon=custom_icon)

        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.label(text="", icon=custom_icon)


class POW_OT_CreateNewPreset(Operator):
    bl_idname = "pow.create_new_preset"
    bl_label = "Create New"
    bl_description = "Create a New Preset from Current POW Scene"

    def invoke(self, context, event):
        # Reset preset name to clear so user doesn't have to delete last preset's name
        context.scene.world.pow_settings.preset_name = ''
        # Create a popup when activated for inputting preset name
        return context.window_manager.invoke_props_dialog(self, width=320)

    def draw(self, context):
        # UI of the popup
        layout = self.layout
        layout.prop(context.scene.world.pow_settings, "preset_name")

    def execute(self, context):
        world = context.scene.world
        invalid_symbols = validate_preset_name(world.pow_settings.preset_name)
        if invalid_symbols:
            # Invalid symbol inputted, do not create a preset and report error
            self.report({'ERROR'}, "Couldn't create the preset. %s" % invalid_symbols[0])
            return {'CANCELLED'}
        else:
            # Find the target directory for our presets and save the JSON there
            base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
            os.makedirs(base_path, exist_ok=True)
            full_path = os.path.join(base_path, world.pow_settings.preset_name + ".json")
            # Check if a preset of such name already exists and cancel operator to prevent overwrite if it's the case
            if os.path.isfile(full_path):
                self.report({'ERROR'}, "Couldn't create a new preset because a preset with provided name (%s) already exists." % world.pow_settings.preset_name)
                return {'CANCELLED'}
            else:
                create_preset(context, full_path)
                update_preset_list(self, context)
                # Set the active preset index to the newly created one
                for i in range(len(world.pow_settings.presets)):
                    if world.pow_settings.presets[i].name == world.pow_settings.preset_name:
                        world.pow_settings.preset_index = i
                        break
                return {'FINISHED'}


class POW_OT_DeleteCurrentPreset(Operator):
    bl_idname = "pow.delete_current_preset"
    bl_label = "Delete Preset"
    bl_description = "Delete the Selected Preset"

    def invoke(self, context, event):
        # Create a confirmation popup to reduce chance of accidental deletion
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        world = context.scene.world
        # Find the target directory for our presets and delete it
        base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
        #os.makedirs(base_path, exist_ok=True)
        full_path = os.path.join(base_path, world.pow_settings.presets[world.pow_settings.preset_index].name + ".json")
        os.remove(full_path)
        # Set the selected preset to default
        world.pow_settings.preset_index = -1
        update_preset_list(self, context)
        return {'FINISHED'}
    

class POW_OT_SaveCurrentPreset(Operator):
    bl_idname = "pow.save_current_preset"
    bl_label = "Save Preset"
    bl_description = "Save the Selected Preset"

    def invoke(self, context, event):
        # Create a confirmation popup to reduce chance of accidental overwrite
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        world = context.scene.world
        # Find the target directory for our presets and save the JSON there
        base_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
        full_path = os.path.join(base_path, world.pow_settings.presets[world.pow_settings.preset_index].name + ".json")
        create_preset(context, full_path)
        update_preset_list(self, context)
        return {'FINISHED'}


class POW_OT_OpenPresetsFolder(Operator):
    bl_idname = "pow.open_presets_folder"
    bl_label = "Open Presets Folder"
    bl_description = "Open the preset folder in file manager"

    def execute(self, context):
        file_path = os.path.join(bpy.utils.user_resource('DATAFILES'), "presets", "POW")
        if sys.platform == "win32":
            os.startfile(file_path)
        else:
            opener = "open" if sys.platform == "darwin" else "xdg-open"
            subprocess.call([opener, file_path])
        return {'FINISHED'}


class POW_OT_ApplyUnderwaterAll(Operator):
    bl_idname = "pow.apply_underwater_all"
    bl_label = "Apply to All Materials"
    bl_description = "Apply underwater nodes to all materials"

    def invoke(self, context, event):
        # Create a confirmation popup to reduce chance of accidental activation
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        update_underwater(self, context, add=True, all_materials=True)

        return {'FINISHED'}


class POW_OT_ApplyUnderwaterSelected(Operator):
    bl_idname = "pow.apply_underwater_selected"
    bl_label = "Apply to Selected Objects"
    bl_description = "Apply underwater nodes to the materials of selected objects"

    def execute(self, context):
        update_underwater(self, context, add=True, all_materials=False)

        return {'FINISHED'}


class POW_OT_RemoveUnderwaterAll(Operator):
    bl_idname = "pow.remove_underwater_all"
    bl_label = "Remove from All Materials"
    bl_description = "Remove underwater nodes from all materials"

    def invoke(self, context, event):
        # Create a confirmation popup to reduce chance of accidental activation
        return context.window_manager.invoke_confirm(self, event)

    def execute(self, context):
        update_underwater(self, context, add=False, all_materials=True)

        return {'FINISHED'}
    

class POW_OT_RemoveUnderwaterSelected(Operator):
    bl_idname = "pow.remove_underwater_selected"
    bl_label = "Remove from Selected Objects"
    bl_description = "Remove underwater nodes from the materials of selected objects"

    def execute(self, context):
        update_underwater(self, context, add=False, all_materials=False)

        return {'FINISHED'}


class POW_OT_AddIntersection(Operator):
    bl_idname = "pow.add_intersection"
    bl_label = "Add Intersection"
    bl_description = "Apply advanced intersection calculation to given objects"

    def execute(self, context):
        update_intersection_nodes(self, context, enable=True)

        return {'FINISHED'}


class POW_OT_RemoveIntersection(Operator):
    bl_idname = "pow.remove_intersection"
    bl_label = "Remove Intersection"
    bl_description = "Remove advanced intersection calculation from given objects"

    def execute(self, context):
        update_intersection_nodes(self, context, enable=False)

        return {'FINISHED'}
